/*---------------------------------------------
S I M P A C K
Simulation Tool Package
Copyright 1990, Paul A. Fishwick
---------------------------------------------*/

#include <stdio.h>
#include "queuing.h"

/* Parameters */
#define TASKS_0 6                 /* no. class 0 tasks  */
#define TASKS_1 3                 /* no. class 1 tasks  */
#define TASKS   TASKS_0+TASKS_1   /* total no. of tasks */
#define DISKS   4                 /* no. of disk units  */

/* Events */
#define BEGIN_TOUR   1
#define REQUEST_CPU  2
#define RELEASE_CPU  3
#define REQUEST_DISK 4
#define RELEASE_DISK 5

struct token
{
	int cls;           /* task's class (& priority) */
	int disk;          /* disk for current IO req.  */
	double tourStartT; /* tour start time stamp     */
} task[TASKS+1];

TOKEN a_token;

int
  disk[DISKS+1],     /* disk facility descriptors */
  cpu,               /* cpu facility descriptor   */
  tours = 500;       /* no. of tours to simulate  */

double
  cpuTime[2] = {10.0, 5.0},           /* class 0,1 mean cpu times  */
  diskTime = 30.0, diskStdDev = 2.5;  /* disk time mean, std. dev. */

main()
{
	int i, j, event, myTours[2];
	double t, myTimes[2], reqTime;
	struct token *p;

	myTours[0] = myTours[1] = 0;
	myTimes[0] = myTimes[1] = 0.0;

	/* set the priroty of the tasks */
  for (i=1; i<=TASKS; i++)
		task[i].cls = (i>TASKS_0)? 1:0;

	/*   init_simpack(LINKED); */
	init_simpack(LINKED);

  /* Create the facilities */
	cpu = create_facility("CPU", 1);

	for (i=1; i<=DISKS; i++)
		disk[i] = create_facility("disk", 1);

  /* Schedule seed events */
	for (i=1; i<=TASKS; i++) {
		a_token.attr[0] = i;
		schedule(BEGIN_TOUR, 0, a_token);
	}

	while (tours) {
		next_event(&event, &a_token);
		i = a_token.attr[0];
		p = &task[i];

		switch(event) {
		case BEGIN_TOUR:  /* begin tour */
			a_token.attr[0] = i;
			p->tourStartT = sim_time();
			schedule(REQUEST_CPU, 0, a_token);
			update_arrivals();
			break;

		case REQUEST_CPU:  /* request cpu */
			j = p->cls;
			a_token.attr[0] = i;
			if (request(cpu, a_token, j) == FREE) {
				reqTime = expntl(cpuTime[j]);
				a_token.attr[0] = i;
				schedule(RELEASE_CPU, (TIME) reqTime, a_token);
			}
			break;

		case RELEASE_CPU:  /* release cpu, select disk */
			a_token.attr[0] = i;
			release(cpu, a_token);
			p->disk = random_gen(1, DISKS);
			schedule(REQUEST_DISK, 0, a_token);
			break;

		case REQUEST_DISK: /* request disk */
			a_token.attr[0]	= i;
			if (request(disk[p->disk], a_token, 0) == FREE){
				reqTime = erlang(diskTime, diskStdDev);
				a_token.attr[0]	= i;
				schedule(RELEASE_DISK, (TIME) reqTime, a_token);
			}
			break;

		case RELEASE_DISK:  /* release disk, end tour */
			a_token.attr[0] = i;
			release(disk[p->disk], a_token);
			j = p->cls;
			t = sim_time();
			myTimes[j] += t-p->tourStartT;
			p->tourStartT = t;
			myTours[j]++;
			update_completions();
			a_token.attr[0] = i;
			schedule(BEGIN_TOUR, 0, a_token);
			tours--;
			break;
		}
	}
	report_stats();
	printf("\n\n");
	printf("class 0 tour time = %.2f\n", myTimes[0]/myTours[0]);
	printf("class 1 tour time = %.2f\n", myTimes[1]/myTours[1]);
}
